#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" General checks & checklists. """

import time
import json
from typing import List
from dessia_common.abstract import CoreDessiaObject


LEVEL_TO_INT = {'debug': 0, 'info': 1, 'warning': 2, 'error': 3}


class PassedCheck:
    """ Denote the result of a check that has no error. """

    level = 'info'

    def __init__(self, message: str):
        self.message = message

    def __repr__(self):
        return f'[{self.level.upper()}: {self.__class__.__name__}] - {self.message}'

    def __bool__(self):
        return True

    def to_dict(self):
        """ Write check as a dict. Used for frontend display. """
        return {"level": self.level, "message": self.message, "object_class": self.__class__.__name__}


class CheckWarning(PassedCheck):
    """ Denote a check warning. """

    level = 'warning'


class FailedCheck(PassedCheck):
    """ Denote a failed check. """

    level = 'error'

    def __bool__(self):
        return False


class BadType(FailedCheck):
    """ Denote a failed check due to a bad type. """


class GeometricInconsistance(FailedCheck):
    """ Denote a failed check due to a geometric inconsistency. """


class NonSerializable(FailedCheck):
    """ Used when a class or an instance is not serializable. """


class NonCopyable(FailedCheck):
    """ Used when an instance can not be copied. """


class NonDisplayable(FailedCheck):
    """ Used when an instance can not be displayed. """


class BadStructure(FailedCheck):
    """ Used when a class or method is badly designed (regarding schemas for example). """


class CheckList:
    """ A list of checks result. """

    def __init__(self, checks: List[PassedCheck]):
        self.checks = checks

    def __repr__(self):
        rep = f'Check list containing {len(self.checks)} checks:\n'
        for check_idx, check in enumerate(self.checks):
            rep += f'Check {check_idx+1} {check}\n'
        return rep

    def __add__(self, other_checklist: 'CheckList'):
        return self.__class__(self.checks + other_checklist.checks)

    def __len__(self):
        return len(self.checks)

    def __getitem__(self, item: int):
        return self.checks[item]

    def checks_above_level(self, level: str = 'error'):
        """ Return True if no check has a level above given one, else False. """
        checks = []
        for check in self.checks:
            if LEVEL_TO_INT[check.level] >= LEVEL_TO_INT[level]:
                checks.append(check)
        return checks

    def raise_if_above_level(self, level: str = 'error'):
        """ Raise an error if some checks have a level above given one. """
        for check in self.checks_above_level(level=level):
            raise ValueError(f'Check: {check} is above level "{level}"')

    def to_dict(self):
        """ Write CheckList as a dict. Used for frontend display. """
        return {"checks": [c.to_dict() for c in self.checks]}


def is_int(value, level: str = 'error'):
    """ Return if value is a int. """
    if not isinstance(value, int):
        return CheckList([FailedCheck(f'Value {value} is not an int')])
    if level == 'info':
        return CheckList([PassedCheck(f'value {value} is an int')])
    return CheckList([])


def is_float(value, level: str = 'error'):
    """ Return if value is a float. """
    if not isinstance(value, float):
        return CheckList([FailedCheck(f'Value {value} is not a float')])
    if level == 'info':
        return CheckList([PassedCheck(f'value {value} is a float')])
    return CheckList([])


def is_str(value, level: str = 'error'):
    """ Return if value is a str. """
    if not isinstance(value, str):
        return CheckList([FailedCheck(f'Value {value} is not a str')])
    if level == 'info':
        return CheckList([PassedCheck(f'value {value} is a str')])
    return CheckList([])


def type_check(value, expected_type, level: str = 'error'):
    """
    Type check the value against the expected type.

    This is experimental!
    """
    if expected_type is int:
        return is_int(value)
    if expected_type is float:
        return is_float(value)
    if expected_type is str:
        return is_str(value)

    if not isinstance(value, expected_type):
        return CheckList([FailedCheck(f'Value {value} is not of type {expected_type}')])
    if level == 'info':
        return CheckList([PassedCheck(f'value {value} is of expected type {expected_type}')])

    return CheckList([])


def check_serialization_process(object_: CoreDessiaObject, use_pointers: bool = True):
    """ Simulate platform serialization/deserialization process to guarantee viability of given object. """
    print("Checking serialization process...")
    start = time.time()
    try:
        dict_ = object_.to_dict(use_pointers=use_pointers)
        print("Serialized. Deserializing...")
    except TypeError as exc:
        if use_pointers:
            # Trying without pointers if it failed with.
            dict_ = object_.to_dict(use_pointers=False)
            print("Serialized without pointers. Deserializing...")
        else:
            print("Failed.\n")
            raise exc

    json_dict = json.dumps(dict_)
    decoded_json = json.loads(json_dict)
    deserialized_object = object_.dict_to_object(decoded_json)
    print("Deserialized. Checking equality...")

    if not deserialized_object._data_eq(object_):
        print('Failed.\nData Diff: ', object_._data_diff(deserialized_object))
        check = NonSerializable('Object is not equal to itself after serialization/deserialization')
    else:
        check = PassedCheck("Object is serializable")
    duration = time.time() - start
    if check:
        print(f"Checked serialization process in {duration}s.\n")
    return {"check": check, "duration": duration, "dict_": dict_}


def check_copy(object_: CoreDessiaObject):
    """ Simulate the copy performed by the platform. """
    print("Copying...")
    start = time.time()
    copied_object = object_.copy()
    print("Copied. Checking equality...")
    if not copied_object._data_eq(object_):
        print("Failed.\n")
        try:
            print('Data Diff: ', object_._data_diff(copied_object))
        except Exception:
            pass
        check = FailedCheck('Object is not equal to itself after copy.')
    else:
        check = PassedCheck("Object can be copied and equal to itself after copy.")
    duration = time.time() - start
    if check:
        print(f"Checked copying process in {duration}s.\n")
    return {"check": check, "duration": duration}


def check_displays(object_: CoreDessiaObject):
    """ Simulate displays computation performed by the platform. """
    print("Checking displays...")
    start = time.time()
    try:
        displays = object_._displays()
        print("Computed displays. Serializing displays...")
    except Exception:
        print("Failed.\n")
        return {"check": NonDisplayable("Object displays failed."), "duration": time.time() - start}

    try:
        json.dumps(displays)
        print("Serialized displays.")
    except Exception:
        print("Failed.\n")
        return {"check": NonSerializable("Object displays serialization failed."), "duration": time.time() - start}
    duration = time.time() - start
    print(f"Checked displays computation process in {duration}s.\n")
    return {"check": PassedCheck("Object displays checks succeeded"), "duration": duration}


def check_schemas(object_: CoreDessiaObject):
    """ Simulate schemas computation performed by the platform. """
    print("Checking schemas...")
    start = time.time()
    try:
        schemas = object_.method_schemas
        print("Computed schemas. Serializing schemas...")
    except Exception as exc:
        print("Failed.\n")
        return {"check": BadStructure(f"Schemas computation failed with following error : \n\n{exc}\n\n"),
                "duration": time.time() - start}

    try:
        json.dumps(schemas)
        print("Serialized schemas.")
    except Exception:
        print("Failed.\n")
        return {"check": NonSerializable("Schemas serialization failed."), "duration": time.time() - start}
    duration = time.time() - start
    print(f"Checked schemas computation process in {duration}s.\n")
    return {"check": PassedCheck("Object displays checks succeeded"), "duration": duration}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Module defining recognized  file types in DC. """
import io
import openpyxl


class BinaryFile(io.BytesIO):
    """ Base class for handling binary files with name. """

    extension = ''

    def __init__(self, filename: str = ''):
        super().__init__()
        self.filename = filename

    def copy(self):
        """ Files deep copy. """
        file_copy = self.__class__(self.filename)
        file_copy.write(self.getbuffer())
        file_copy.seek(0)
        return file_copy

    @classmethod
    def stream_template(cls):
        """ Get template of class of file. """
        template = cls()
        return template

    @classmethod
    def from_file(cls, filepath: str):
        """ Get a file from a binary file. """
        with open(filepath, 'rb') as file:
            stream = cls()
            stream.write(file.read())
            stream.seek(0)
            return stream

    @classmethod
    def save_template_to_file(cls, filename):
        """ Save template of class into a file. """
        if cls.extension and not filename.endswith(cls.extension):
            filename = f'{filename}.{cls.extension}'
        with open(filename, 'wb') as file:
            stream = cls.stream_template()
            stream.seek(0)
            file.write(stream.getvalue())

    def __hash__(self):
        return hash(self.filename)

    def __eq__(self, other):
        return isinstance(other, BinaryFile) and self.getbuffer() == other.getbuffer() \
               and self.filename == other.filename


class StringFile(io.StringIO):
    """ Base class for handling text files with name with default encoding : utf-8. """

    extension = ''

    def __init__(self, filename: str = ''):
        super().__init__()
        self.filename = filename

    def copy(self):
        """ Files deep copy. """
        file_copy = self.__class__(self.filename)
        file_copy.write(self.getvalue())
        file_copy.seek(0)
        return file_copy

    @classmethod
    def stream_template(cls):
        """ Get template of class of file. """
        template = cls()
        template.write('Text file template. Subclass the class dessia_common.files.StringFile'
                       ' to define your own filetype')
        template.seek(0)
        return template

    @classmethod
    def from_stream(cls, stream_file):
        """ Get a file from a stream file. """
        stream = StringFile()
        stream.write(stream_file.read().decode('utf-8'))
        stream.seek(0)
        return stream

    @classmethod
    def from_file(cls, filepath: str):
        """ Get a file from a file. """
        with open(filepath, 'r', encoding='utf-8') as file:
            stream = cls()
            stream.write(file.read())
            stream.seek(0)
            return stream

    @classmethod
    def save_template_to_file(cls, filename):
        """ Save instantiated template into a file. """
        if cls.extension and not filename.endswith(cls.extension):
            filename = f'{filename}.{cls.extension}'
        with open(filename, 'w', encoding='utf-8') as file:
            stream = cls.stream_template()
            stream.seek(0)
            file.write(stream.getvalue())

    def __hash__(self):
        return hash(self.filename)

    def __eq__(self, other):
        return isinstance(other, StringFile) and self.getvalue() == other.getvalue() \
               and self.filename == other.filename


class XLSXFile(BinaryFile):
    """ Base class for Excel XLS files. """

    extension = 'xlsx'

    @classmethod
    def stream_template(cls):
        """ Get template of class of file. """
        template = cls()
        workbook = openpyxl.Workbook()
        sheet = workbook.active
        sheet.title = 'Template of XLSX'
        sheet.cell(1, 1).value = 'Subclass this class to define your own template'
        workbook.save(template)
        template.seek(0)
        return template


class XLSFile(BinaryFile):
    """ Base class for Excel files in old format. """

    extension = 'xls'


class XLSMFile(XLSXFile):
    """ Base class for Excel XLS files with macros. """

    extension = 'xlsm'


class TextFile(StringFile):
    """ Base class for text files. """

    extension = 'txt'


class CSVFile(StringFile):
    """
    Base class for Coma Separated Values (CSV) files.

    https://en.wikipedia.org/wiki/Comma-separated_values
    """

    extension = 'csv'

    @classmethod
    def stream_template(cls):
        """ Get template of class of file. """
        template = cls()
        template.write('"col A", "col B", "col C"\n')
        template.write('"abc", 1, 1.23\n')

        template.seek(0)
        return template


class MarkdownFile(StringFile):
    """
    Base class for markdown files.

    https://en.wikipedia.org/wiki/Markdown
    """

    extension = 'md'

    @classmethod
    def stream_template(cls):
        """ Get template of class of file. """
        template = cls()
        template.write('# Text file template.\n\n'
                       'Subclass the class dessia_common.files.StringFile'
                       ' to define your own filetype')
        template.seek(0)
        return template


class JsonFile(StringFile):
    """ A .json extended file. """

    extension = "json"


class DocFile(BinaryFile):
    """ Base class for Word files in old format. """

    extension = 'doc'


class DocxFile(BinaryFile):
    """ Base class for Word files in new XML format. """

    extension = 'docx'


def generate_archive(zip_archive, value):
    """ Write the contents of a file-like object to a ZipFile archive. """
    if isinstance(value, StringFile):
        with zip_archive.open(value.filename, 'w') as file:
            file.write(value.getvalue().encode('utf-8'))
    elif isinstance(value, BinaryFile):
        with zip_archive.open(value.filename, 'w') as file:
            file.write(value.getbuffer())
    else:
        raise ValueError(f"Archive input is not a file-like object. Got '{value}' of type {type(value)}")
    return file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" JsonSchema generation functions. """
import warnings
import dessia_common.utils.types as dc_types

JSONSCHEMA_HEADER = {"definitions": {},
                     "$schema": "http://json-schema.org/draft-07/schema#",
                     "type": "object",
                     "required": [],
                     "properties": {}}

TYPING_EQUIVALENCES = {int: 'number', float: 'number', bool: 'boolean', str: 'string'}


def default_sequence(array_jsonschema):
    """ DEPRECATED. Soon to be removed. """
    warnings.warn("Jsonschema module is deprecated and will be removed soon. Use schemas instead.", DeprecationWarning)
    if dc_types.is_sequence(array_jsonschema['items']):
        # Tuple jsonschema
        if 'default_value' in array_jsonschema:
            return array_jsonschema['default_value']
        return [default_dict(v) for v in array_jsonschema['items']]
    return None


def datatype_from_jsonschema(jsonschema):
    """ DEPRECATED. Soon to be removed. """
    warnings.warn("Jsonschema module is deprecated and will be removed soon. Use schemas instead.", DeprecationWarning)
    if jsonschema['type'] == 'object':
        if 'classes' in jsonschema:
            if len(jsonschema['classes']) > 1:
                return 'union'
            if 'standalone_in_db' in jsonschema:
                if jsonschema['standalone_in_db']:
                    return 'standalone_object'
                return 'embedded_object'
            # Static dict is deprecated
            return 'static_dict'
        if 'instance_of' in jsonschema:
            return 'instance_of'
        if 'patternProperties' in jsonschema:
            return 'dynamic_dict'
        if 'is_method' in jsonschema and jsonschema['is_method']:
            return 'embedded_object'
        if 'is_class' in jsonschema and jsonschema['is_class']:
            return 'class'

    if jsonschema['type'] == 'array':
        if 'additionalItems' in jsonschema and not jsonschema['additionalItems']:
            return 'heterogeneous_sequence'
        return 'homogeneous_sequence'

    if jsonschema["type"] == "text" and "is_file" in jsonschema and jsonschema["is_file"]:
        return "file"

    if jsonschema['type'] in ['number', 'string', 'boolean']:
        return 'builtin'
    return None


def chose_default(jsonschema):
    """ DEPRECATED. Soon to be removed. """
    warnings.warn("Jsonschema module is deprecated and will be removed soon. Use schemas instead.", DeprecationWarning)
    datatype = datatype_from_jsonschema(jsonschema)
    if datatype in ['heterogeneous_sequence', 'homogeneous_sequence']:
        return default_sequence(jsonschema)
    if datatype == 'static_dict':
        # Deprecated
        return default_dict(jsonschema)
    if datatype in ['standalone_object', 'embedded_object', 'instance_of', 'union']:
        if 'default_value' in jsonschema:
            return jsonschema['default_value']
        return None

    return None


def default_dict(jsonschema):
    """ DEPRECATED. Soon to be removed. """
    warnings.warn("Jsonschema module is deprecated and will be removed soon. Use schemas instead.", DeprecationWarning)
    dict_ = {}
    datatype = datatype_from_jsonschema(jsonschema)
    if datatype in ['standalone_object', 'embedded_object', 'static_dict']:
        if 'classes' in jsonschema:
            dict_['object_class'] = jsonschema['classes'][0]
        elif 'is_method' in jsonschema and jsonschema['is_method']:
            # Method can have no classes in jsonschema
            pass
        else:
            raise ValueError(f"DessiaObject of type {jsonschema['python_typing']} must have 'classes' in jsonschema")
        for property_, jss in jsonschema['properties'].items():
            if 'default_value' in jss:
                dict_[property_] = jss['default_value']
            else:
                dict_[property_] = chose_default(jss)
    else:
        return None
    return dict_



"""
Type Checking interfaces for schemas.

The aim of this module is to provide an "API documentation" or at least to state the body of schemas
in order to define a language understandable by both frontend and dessia_common.
It heavily relies on data typing and propose ways to type function return types, for example.
We could test that dessia_common is able to write structures validating these schemas
and ensure that frontend is able to read them.
By keeping them both side, it would decrease the risk of regression issues

This is mostly experimental and might require higher versions of python in order to work properly.
As an example, python >= 3.11 could provide a way to define required or not required attribute
in generated dictionaries.

At the very moment these lines are written (25 May 2023), Pycharm is complaining over dictionary specification.
The base dictionary generated for a Property validates its schema. However, when "extending" it to OptionalProperty,
for instance, it complains about type casting, and it seems that no solutions are currently available.
"""
import typing as tp


T = tp.TypeVar("T")

Annotations = tp.Dict[str, tp.Type[T]]


class PropertySchema(tp.TypedDict):
    """ Typing for properties. Experimental. """

    title: str
    editable: bool
    description: str
    python_typing: str
    type: str


class OptionalPropertySchema(PropertySchema):
    """ Typing for Optional properties. Experimental. """

    default_value: tp.Any


class TupleSchema(PropertySchema):
    """ Typing for Tuple Schemas. Experimental. """

    additionalItems: bool
    items: tp.List[PropertySchema]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb  3 14:10:54 2022

@author: steven
"""

import io
import importlib
import inspect

module = importlib.import_module('dessia_common.files')

for name, obj in inspect.getmembers(module):
    if inspect.isclass(obj) and issubclass(obj, (io.BytesIO, io.StringIO)):
        template = obj.save_template_to_file('test')
